/// <reference types="node" />
import { Connection, PublicKey } from "@solana/web3.js";
import { Schema } from "borsh";
import BN from "bn.js";
import { AccountTag } from "./market_state";
/** @enum {number} */
export declare enum EventType {
    Fill = 0,
    Out = 1
}
/** @enum {number} */
export declare enum Side {
    Bid = 0,
    Ask = 1
}
/**
 * Event queue header object
 */
export declare class EventQueueHeader {
    tag: AccountTag;
    head: BN;
    count: BN;
    seqNum: BN;
    static LEN: number;
    static schema: Schema;
    constructor(arg: {
        tag: number;
        head: BN;
        count: BN;
        seqNum: BN;
    });
}
/**
 * Event fill object
 */
export declare class EventFill {
    takerSide: Side;
    quoteSize: BN;
    makerOrderId: BN;
    baseSize: BN;
    makerCallbackInfo: number[];
    takerCallbackInfo: number[];
    static LEN: number;
    static schema: Schema;
    constructor(arg: {
        tag: number;
        takerSide: Side;
        quoteSize: BN;
        makerOrderId: BN;
        baseSize: BN;
    });
}
/**
 * EventOut object
 */
export declare class EventOut {
    side: Side;
    orderId: BN;
    baseSize: BN;
    delete: boolean;
    callbackInfo: number[];
    static schema: Schema;
    constructor(arg: {
        side: number;
        orderId: BN;
        baseSize: BN;
        delete: number;
    });
}
/**
 * Event queue object
 */
export declare class EventQueue {
    header: EventQueueHeader;
    eventsBuffer: number[];
    callbackInfosBuffer: number[];
    callBackInfoLen: number;
    constructor(arg: {
        header: EventQueueHeader;
        eventsBuffer: number[];
        callbackInfosBuffer: number[];
        callBackInfoLen: number;
    });
    /**
     * Deserialize a buffer into an EventQueue object
     * @param callBackInfoLen Length of the callback information
     * @param data Buffer to deserialize
     * @returns Returns an EventQueue object
     */
    static parse(callBackInfoLen: number, data: Buffer): EventQueue;
    /**
     * Loads the event queue from its address
     * @param connection The solana connection object to the RPC node
     * @param address The address of the event queue
     * @param callBackInfoLen The length of the callback information
     * @returns Returns an EventQueue object
     */
    static load(connection: Connection, address: PublicKey, callBackInfoLen: number): Promise<EventQueue>;
    /**
     * Returns an event from its index in the event queue
     * @param idx Index of the event to parse
     * @returns Returns an Event object
     */
    parseEvent(idx: number): EventFill | EventOut;
    /**
     * Retrieves the event at position `idx` in the queue.
     * @param idx Index of the event to peek and parse
     * @returns Returns an Event object
     */
    peekAt(idx: number): EventFill | EventOut;
    /**
     * Returns fill events from the event queue
     * @param limit Optional limit parameter
     * @returns An array of EventFill
     */
    parseFill(limit?: number): (EventFill | EventOut)[];
    /**
     * Deserialize a buffer into an EventQueueHeader object
     * @param data Buffer to deserialize
     * @returns Returns an EventQueueHeader object
     */
    static parseEventQueueHeader(data: Buffer): EventQueueHeader;
    static computeAllocationSize(desiredEventCapacity: number, callbackInfoLen: number): number;
}
