"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EventQueue = exports.EventOut = exports.EventFill = exports.EventQueueHeader = exports.Side = exports.EventType = void 0;
const borsh_1 = require("borsh");
/** @enum {number} */
var EventType;
(function (EventType) {
    EventType[EventType["Fill"] = 0] = "Fill";
    EventType[EventType["Out"] = 1] = "Out";
})(EventType = exports.EventType || (exports.EventType = {}));
/** @enum {number} */
var Side;
(function (Side) {
    Side[Side["Bid"] = 0] = "Bid";
    Side[Side["Ask"] = 1] = "Ask";
})(Side = exports.Side || (exports.Side = {}));
/**
 * Event queue header object
 */
class EventQueueHeader {
    constructor(arg) {
        this.tag = arg.tag;
        this.head = arg.head;
        this.count = arg.count;
        this.seqNum = arg.seqNum;
    }
}
exports.EventQueueHeader = EventQueueHeader;
EventQueueHeader.LEN = 32;
EventQueueHeader.schema = new Map([
    [
        EventQueueHeader,
        {
            kind: "struct",
            fields: [
                ["tag", "u64"],
                ["head", "u64"],
                ["count", "u64"],
                ["seqNum", "u64"],
            ],
        },
    ],
]);
/**
 * Event fill object
 */
class EventFill {
    constructor(arg) {
        this.takerSide = arg.takerSide;
        this.makerOrderId = arg.makerOrderId;
        this.quoteSize = arg.quoteSize;
        this.baseSize = arg.baseSize;
    }
}
exports.EventFill = EventFill;
EventFill.LEN = 40;
EventFill.schema = new Map([
    [
        EventFill,
        {
            kind: "struct",
            fields: [
                ["tag", "u8"],
                ["takerSide", "u8"],
                ["_padding", [6]],
                ["quoteSize", "u64"],
                ["makerOrderId", "u128"],
                ["baseSize", "u64"],
            ],
        },
    ],
]);
/**
 * EventOut object
 */
class EventOut {
    constructor(arg) {
        this.side = arg.side;
        this.orderId = arg.orderId;
        this.baseSize = arg.baseSize;
        this.delete = arg.delete === 1;
    }
}
exports.EventOut = EventOut;
EventOut.schema = new Map([
    [
        EventOut,
        {
            kind: "struct",
            fields: [
                ["tag", "u8"],
                ["side", "u8"],
                ["delete", "u8"],
                ["_padding", [13]],
                ["orderId", "u128"],
                ["baseSize", "u64"],
            ],
        },
    ],
]);
/**
 * Event queue object
 */
class EventQueue {
    constructor(arg) {
        this.header = arg.header;
        this.eventsBuffer = arg.eventsBuffer;
        this.callbackInfosBuffer = arg.callbackInfosBuffer;
        this.callBackInfoLen = arg.callBackInfoLen;
    }
    /**
     * Deserialize a buffer into an EventQueue object
     * @param callBackInfoLen Length of the callback information
     * @param data Buffer to deserialize
     * @returns Returns an EventQueue object
     */
    static parse(callBackInfoLen, data) {
        let header = borsh_1.deserializeUnchecked(EventQueueHeader.schema, EventQueueHeader, data);
        let capacity = (data.length - EventQueueHeader.LEN) /
            (EventFill.LEN + 2 * callBackInfoLen);
        let callbackInfosOffset = EventQueueHeader.LEN + capacity * EventFill.LEN;
        let eventsBuffer = data.slice(EventQueueHeader.LEN, callbackInfosOffset);
        let callbackInfosBuffer = data.slice(callbackInfosOffset);
        return new EventQueue({
            header,
            eventsBuffer: [...eventsBuffer],
            callbackInfosBuffer: [...callbackInfosBuffer],
            callBackInfoLen,
        });
    }
    /**
     * Loads the event queue from its address
     * @param connection The solana connection object to the RPC node
     * @param address The address of the event queue
     * @param callBackInfoLen The length of the callback information
     * @returns Returns an EventQueue object
     */
    static async load(connection, address, callBackInfoLen) {
        const accountInfo = await connection.getAccountInfo(address);
        if (!(accountInfo === null || accountInfo === void 0 ? void 0 : accountInfo.data)) {
            throw new Error("Invalid address provided");
        }
        return this.parse(callBackInfoLen, accountInfo.data);
    }
    /**
     * Returns an event from its index in the event queue
     * @param idx Index of the event to parse
     * @returns Returns an Event object
     */
    parseEvent(idx) {
        let eventsOffset = idx * EventFill.LEN;
        let data = Buffer.from(this.eventsBuffer.slice(eventsOffset, eventsOffset + EventFill.LEN));
        switch (data[0]) {
            case EventType.Fill: {
                let event = borsh_1.deserializeUnchecked(EventFill.schema, EventFill, data);
                let makerOffset = 2 * idx * this.callBackInfoLen;
                let takerOffset = (2 * idx + 1) * this.callBackInfoLen;
                event.makerCallbackInfo = this.callbackInfosBuffer.slice(makerOffset, makerOffset + this.callBackInfoLen);
                event.takerCallbackInfo = this.callbackInfosBuffer.slice(takerOffset, takerOffset + this.callBackInfoLen);
                return event;
            }
            case EventType.Out: {
                let event = borsh_1.deserializeUnchecked(EventOut.schema, EventOut, data);
                let offset = 2 * idx * this.callBackInfoLen;
                event.callbackInfo = this.callbackInfosBuffer.slice(offset, offset + this.callBackInfoLen);
                return event;
            }
            default:
                throw new Error("Invalid data provided");
        }
    }
    /**
     * Retrieves the event at position `idx` in the queue.
     * @param idx Index of the event to peek and parse
     * @returns Returns an Event object
     */
    peekAt(idx) {
        const len = this.eventsBuffer.length / EventFill.LEN;
        return this.parseEvent((idx + this.header.head.toNumber()) % len);
    }
    /**
     * Returns fill events from the event queue
     * @param limit Optional limit parameter
     * @returns An array of EventFill
     */
    parseFill(limit) {
        const n = limit
            ? Math.min(limit, this.header.count.toNumber())
            : this.header.count.toNumber();
        return [...Array(n).keys()]
            .map((e) => this.parseEvent(e))
            .filter((e) => e instanceof EventFill);
    }
    /**
     * Deserialize a buffer into an EventQueueHeader object
     * @param data Buffer to deserialize
     * @returns Returns an EventQueueHeader object
     */
    static parseEventQueueHeader(data) {
        return borsh_1.deserialize(EventQueueHeader.schema, EventQueueHeader, data);
    }
    static computeAllocationSize(desiredEventCapacity, callbackInfoLen) {
        return (desiredEventCapacity * (EventFill.LEN + 2 * callbackInfoLen) +
            EventQueueHeader.LEN);
    }
}
exports.EventQueue = EventQueue;
//# sourceMappingURL=event_queue.js.map